from alternat.generation.config import Config
from alternat.generation.rules.caption_handler import CaptionDataHandler
from alternat.generation.rules.label_handler import LabelDataHandler
from alternat.generation.rules.ocr_handler import OCRDataHandler

from .version import __version__




from treelib import Tree
from alternat.generation.exceptions import OCRClusteringException


class ClusterTree(object):
    """Alternat clustering implementation class.

    :param object: [description]
    :type object: [type]
    :return: [description]
    :rtype: [type]
    """

    # the bounding boxes should fall within threshold * 100 of width or height
    OVERLAP_THRESHOLD = 0.95

    def __init__(self):
        self.tree = Tree()
        self.root = "root"
        self.tree.create_node(self.root, self.root, data=[])

    def add_node(self, parent_name: str, node_name: str, data):
        """Adds a node in the tree (uses treelib).

        :param parent_name: Tag of the parent in the tree.
        :type parent_name: str
        :param node_name: Tag of the node to be added in the tree.
        :type node_name: str
        :param data: Data to be stored in the node.
        :type data: [type]
        """
        self.tree.create_node(node_name.upper(), node_name, data=data, parent=parent_name)

    def _find_next_overlapping_line_top(self, lines_sorted_top, current_line, line_with_max_bottom):
        """Find the next overlapping element based on top values of line.

        :param lines_sorted_top: Sorted list of lines based on their top values.
        :type lines_sorted_top: [type]
        :param current_line: Current line object.
        :type current_line: [type]
        :param line_with_max_bottom: Line with max bottom value.
        :type line_with_max_bottom: [type]
        :return: [description]
        :rtype: [type]
        """
        starting_index = lines_sorted_top.index(current_line)
        remaining_lines = lines_sorted_top[starting_index + 1:]

        max_line_height = line_with_max_bottom["bottom"] - line_with_max_bottom["top"]
        threshold_height = max_line_height * self.OVERLAP_THRESHOLD

        overlapping_lines = list(filter(lambda line: (line["top"] >= current_line["top"]) and
                                                     (line["top"] < (line_with_max_bottom["top"] + threshold_height)),
                                        remaining_lines))

        if len(overlapping_lines) > 0:

            next_overlapping_line = overlapping_lines[0]
            new_line_with_max_bottom = line_with_max_bottom

            if next_overlapping_line["bottom"] > line_with_max_bottom["bottom"]:
                new_line_with_max_bottom = next_overlapping_line

            return [next_overlapping_line, new_line_with_max_bottom]
        else:
            return [None, line_with_max_bottom]

    def _find_next_overlapping_line_left(self, lines_sorted_left, current_line, line_with_max_right):
        """Find the next overlapping element based on left values of line.

        :param lines_sorted_left: Sorted list of lines based on their left values.
        :type lines_sorted_left: [type]
        :param current_line: Current line object.
        :type current_line: [type]
        :param line_with_max_right: Line with max right value.
        :type line_with_max_right: [type]
        :return: [description]
        :rtype: [type]
        """
        starting_index = lines_sorted_left.index(current_line)
        remaining_lines = lines_sorted_left[starting_index + 1:]

        max_line_right = line_with_max_right["right"] - line_with_max_right["left"]
        threshold_width = max_line_right * self.OVERLAP_THRESHOLD

        overlapping_lines = list(filter(lambda line: (line["left"] >= current_line["left"]) and
                                                     (line["left"] < (line_with_max_right["left"] + threshold_width)),
                                        remaining_lines))

        if len(overlapping_lines) > 0:

            next_overlapping_line = overlapping_lines[0]
            new_line_with_max_right = line_with_max_right

            if next_overlapping_line["right"] > line_with_max_right["right"]:
                new_line_with_max_right = next_overlapping_line

            return [next_overlapping_line, new_line_with_max_right]
        else:
            return [None, line_with_max_right]

    def _create_row_clusters(self, data: list):
        """Create row clusters based on the transformed line data.

        :param data: Transformed line data with bounding box info in the form {top: value, right: value, bottom: value, left: value}.
        :type data: list
        :return: [description]
        :rtype: [type]
        """
        clusters = []
        cluster_number = 0

        line_index = 0

        sorted_top = sorted(data, key=lambda line: line["top"])

        while line_index < len(data):
            cluster = []
            cluster_number += 1
            line_with_max_bottom = None

            next_line_within_height = sorted_top[line_index]
            line_with_max_bottom = sorted_top[line_index]

            while next_line_within_height is not None:
                cluster.append(next_line_within_height)
                current_line = next_line_within_height

                overlapping_line_data = self._find_next_overlapping_line_top(sorted_top,
                                                                             current_line,
                                                                             line_with_max_bottom)

                next_line_within_height = overlapping_line_data[0]
                line_with_max_bottom = overlapping_line_data[1]

            sorted_cluster = sorted(cluster, key=lambda line: line["left"])

            clusters.append(sorted_cluster)
            line_index += len(cluster)

        return clusters

    def _create_column_clusters(self, data: list):
        """Creates column clusters in the transformed line data.

        :param data: Transformed line data with bounding box info in the form {top: value, right: value, bottom: value, left: value}.
        :type data: list
        :return: [description]
        :rtype: [type]
        """
        clusters = []
        cluster_number = 0

        line_index = 0

        sorted_left = sorted(data, key=lambda line: line["left"])

        while line_index < len(data):
            cluster = []
            cluster_number += 1
            line_with_max_right = None

            next_line_within_width = sorted_left[line_index]
            line_with_max_right = sorted_left[line_index]

            while next_line_within_width is not None:
                cluster.append(next_line_within_width)
                current_line = next_line_within_width

                overlapping_line_data = self._find_next_overlapping_line_left(sorted_left,
                                                                              current_line,
                                                                              line_with_max_right)

                next_line_within_width = overlapping_line_data[0]
                line_with_max_right = overlapping_line_data[1]

            sorted_cluster = sorted(cluster, key=lambda line: line["top"])

            clusters.append(sorted_cluster)
            line_index += len(cluster)

        return clusters

    def _combine_nodes(self, data: list):
        """Combine nodes which are part of the cluster because they are closed to each other

        :param data: Transformed line data with bounding box info in the form {top: value, right: value, bottom: value, left: value}.
        :type data: list
        :return: [description]
        :rtype: [type]
        """
        node_text = ""

        sorted_top = sorted(data, key=lambda line: line["top"])
        for line_data in sorted_top:
            node_text += line_data["text"] + " "

        return {
            "left": min([line_data["left"] for line_data in sorted_top]),
            "top": min([line_data["top"] for line_data in sorted_top]),
            "right": max([line_data["right"] for line_data in sorted_top]),
            "bottom": max([line_data["bottom"] for line_data in sorted_top]),
            "text": node_text
        }

    def _create_tree(self, cluster_type: str, lines_data: list, parent_name: str, parent_cluster_type: str = None):
        """Recursive method to cluster lines data into columns and rows clusters. The recursion return when the leaf
        node is encountered.

        :param cluster_type: The type cluster to make (ROW / COL).
        :type cluster_type: str
        :param lines_data: Transformed line data with bounding box info in the form {top: value, right: value, bottom: value, left: value}.
        :type lines_data: list
        :param parent_name: Name / Tag of the parent node.
        :type parent_name: str
        :param parent_cluster_type: The type of cluster for parent, defaults to None (ROW)
        :type parent_cluster_type: str, optional
        """

        row_index = 0
        column_index = 0

        if cluster_type == "ROW":

            if len(lines_data) > 1:

                cluster_data = self._create_row_clusters(lines_data)

                if parent_cluster_type == "COL" and len(cluster_data) == 1:
                    # add a leaf node with combined data
                    node = self._combine_nodes(cluster_data[0])

                    node_name = parent_name + "L"
                    self.add_node(parent_name, node_name, {
                        "children": [],
                        "val": node
                    })
                else:
                    for data in cluster_data:
                        row_index += 1
                        node_name = parent_name + "R" + str(row_index)
                        self.add_node(parent_name, node_name, {
                            "children": data
                        })

                        # print("Row cluster : ", row_index, data, node_name)

                        self._create_tree("COL", data, node_name, "ROW")
            else:

                if len(lines_data) == 1:
                    node_name = parent_name + "L"
                    self.add_node(parent_name, node_name, {
                        "children": [],
                        "val": lines_data[0]
                    })

        if cluster_type == "COL":
            if len(lines_data) > 1:

                cluster_data = self._create_column_clusters(lines_data)

                if parent_cluster_type == "ROW" and len(cluster_data) == 1:
                    # add a leaf node with combined data
                    node = self._combine_nodes(cluster_data[0])

                    node_name = parent_name + "L"
                    self.add_node(parent_name, node_name, {
                        "children": [],
                        "val": node
                    })

                else:
                    for data in cluster_data:
                        column_index += 1
                        node_name = parent_name + "C" + str(column_index)
                        self.add_node(parent_name, node_name, {
                            "children": data
                        })

                        # print("Column cluster : ", column_index, data, node_name)

                        self._create_tree("ROW", data, node_name, "COL")
            else:
                if len(lines_data) == 1:
                    node_name = parent_name + "L"
                    self.add_node(parent_name, node_name, {
                        "children": [],
                        "val": lines_data[0]
                    })

    @staticmethod
    def _update_data(lines_data: list):
        """Transforms ocr bounding box data from rule engine of form 
        [{x: left, y: top}, {x: right, y: top}, {x: right, y: bottom}, {x: left, y: bototm}] to [{top: value, right: value, left: value, bottom: value}]
        :param lines_data: OCR data after passing from rule engine.
        :type lines_data: list
        :return: [description]
        :rtype: [type]
        """
        updated_lines_data = []
        for line in lines_data:

            left = min([bb["x"] for bb in line["boundingBox"]])
            top = min([bb["y"] for bb in line["boundingBox"]])
            right = max([bb["x"] for bb in line["boundingBox"]])
            bottom = max([bb["y"] for bb in line["boundingBox"]])
            new_data = {
                "top": top,
                "left": left,
                "right": right,
                "bottom": bottom,
                "confidence": line["confidence"],
                "text": line["text"]
            }

            updated_lines_data.append(new_data)

        return updated_lines_data

    def _generate_ocr_text_from_tree(self):
        """Generates OCR text from the tree. Does the Depth-first traversal
        to find leaf nodes and appends the text to generate final OCR text.

        :return: [description]
        :rtype: [type]
        """
        ocr_text = ""
        for node in self.tree.expand_tree(mode=Tree.DEPTH):
            node_id = self.tree[node].tag
            last_char_node_id = node_id[len(node_id) - 1]

            # only if the node is the leaf node
            if last_char_node_id == "L":
                # print("Running for node ", node_id, self.tree[node].data["val"]["text"])
                ocr_text += self.tree[node].data["val"]["text"] + " "
        return ocr_text

    def generate_data(self, lines_data):
        """Main handler for alternat clustering implementation. The method
        internally creates clusters, save it on the tree and generates the final
        OCR text from it.

        :param lines_data: OCR data after passing from rule engine.
        :type lines_data: [type]
        :return: [description]
        :rtype: [type]
        """
        modified_lines_data = self._update_data(lines_data)
        #
        # print(" line data : ", modified_lines_data)
        cluster_type = "ROW"
        parent_name = self.root

        # adding a general exception to catch infinite loop error
        try:
            self._create_tree(cluster_type, modified_lines_data, parent_name)
        except:
            OCRClusteringException()
            return ""

        ocr_data = self._generate_ocr_text_from_tree()
        return ocr_data













from alternat.generation.base.config import AnalyzerConf


class Config(AnalyzerConf):
    """Configuration class for Azure driver.

    :param AnalyzerConf: Base configuration class for driver.
    :type AnalyzerConf: [type]
    :return: [description]
    :rtype: [type]
    """
    SUBSCRIPTION_KEY = "" # ENTER AZURE COGNITIVE SERVICE SUBSCRIPTION KEY HERE
    ENDPOINT = "https://<ENTER_PROJECT_NAME>.cognitiveservices.azure.com/" # ENTER ENDPOINT FOR COGNITIVE SERVICE API

    # set this to true if rate limit needs to be enabled
    AZURE_RATE_LIMIT_ON = True

    # The sleep time after which the next set of requests are send
    AZURE_RATE_LIMIT_TIME_IN_SEC = 30

    # the API wont process if height or width width of the image go outside of
    # these constraints
    _MIN_IMAGE_SIZE_IN_PIXEL = 50
    _MAX_IMAGE_SIZE_IN_PIXEL = 5000

    # driver specific threshold for caption (used by rule engine)
    CAPTION_CONFIDENCE_THRESHOLD = 0.20

    # driver specific threshold for OCR (used by rule engine)
    OCR_CONFIDENCE_THRESHOLD = 0.75

    # driver specific threshold for label data (used by rule engine)
    LABEL_CONFIDENCE_THRESHOLD = 0.75

    # line height to image height ratio for OCR
    OCR_HEIGHT_RATIO_TO_IMAGE_THRESHOLD = 0.015

    @classmethod
    def params(cls) -> dict:
        """Retreives the subscription key and the endpoint URL in the JSON.

        :return: [description]
        :rtype: dict
        """
        return {
            "subscription": cls.SUBSCRIPTION_KEY,
            "endpoint": cls.ENDPOINT,
        }

    def update_conf(self, conf: dict):
        """Update configuration based on parameters passed in the JSON. The class members 
        and the key in the conf JSON should match.

        :param conf: JSON object with configuration.
        :type conf: dict
        """
        self.AZURE_RATE_LIMIT_ON = conf["AZURE_RATE_LIMIT_ON"]



import shutil
import os
import subprocess
import sys
import pathlib


class Collector:

    def __init__(self):
        """
        """

    def process(
        self,
        url: str,
        output_dir_path: str,
        download_recursive: bool = False,
        collect_using_apify: bool = False,
    ):
        """Collects image from the url into the output directory

        :param url: [description]
        :type url: str
        :param output_dir_path: [description]
        :type output_dir_path: str
        :param download_recursive: [description], defaults to False
        :type download_recursive: bool, optional
        :param collect_using_apify: [description], defaults to False
        :type collect_using_apify: bool, optional
        """
        startURLArg = "start_urls=" + url
        try:
            os.environ["DOWNLOAD_URL"] = url
            os.environ["OUTPUT_FOLDER"] = output_dir_path
            if not os.path.exists(output_dir_path):
                os.makedirs(output_dir_path)
            if not collect_using_apify:
                print("No crawler explicitly mentioned, using default apify crawler")
                collect_using_apify = True
            if collect_using_apify:
                # Get directory name
                apify_storage_dir = os.path.join(".", "apify_storage")
                current_dir = pathlib.Path(__file__).parent.absolute()
                node_folder_path = os.path.join(os.path.expanduser("~"), ".alternat", "node_modules")
                os.environ["NODE_PATH"] = node_folder_path
                os.environ["APIFY_LOCAL_STORAGE_DIR"] = apify_storage_dir
                ## Try to remove tree; if failed show an error using try...except on screen
                try:
                    shutil.rmtree(apify_storage_dir)
                except OSError as e:
                    print ("Error: %s - %s." % (e.filename, e.strerror))
                if download_recursive:
                    # TODO remove slash to os.path
                    # TODO remove path to current file
                    script_path = os.path.join(current_dir, "apify", "main.js")
                    output = subprocess.check_output(["node", script_path,])
                else:
                    print(f"Downloading single page", url)
                    script_path = os.path.join(current_dir, "apify", "main_single_page.js")
                    output = subprocess.check_output(["node", script_path,])
                # TODO Remove print statement 
                #print("output", output, flush=True)
            else:
                print("Set at least collect-using-apify or collect-using-scrapy to True")

        except subprocess.CalledProcessError as e:
            print("\n Subprocess error")
            print(str(e.output))
        except Exception as e:
            print("\n\n\nException occurred doing cleanup\n\n\n")
            print(e)
            raise



import os
import json
import pathlib
from alternat.generation.base.config import AnalyzerConf


class Config(AnalyzerConf):
    """Configuration class for Google driver.

    :param AnalyzerConf: Base configuration class for driver.
    :type AnalyzerConf: [type]
    :return: [description]
    :rtype: [type]
    """

    # The filename of the credentials file
    _GOOGLE_APPLICATION_CREDENTIALS_FILENAME = 'key.json'

    # The relative path to directory in which the credentials file should exist
    _RELATIVE_CREDENTIALS_DIR_PATH = "credentials"

     # driver specific threshold for caption (used by rule engine)
    CAPTION_CONFIDENCE_THRESHOLD = 0.00

    # driver specific threshold for OCR (used by rule engine)
    OCR_CONFIDENCE_THRESHOLD = 0.70

    # driver specific threshold for label data (used by rule engine)
    LABEL_CONFIDENCE_THRESHOLD = 0.50

    # line height to image height ratio for OCR
    OCR_HEIGHT_RATIO_TO_IMAGE_THRESHOLD = 0.015

    # absolute path to credential file
    ABSOLUTE_PATH_TO_CREDENTIAL_FILE = ""

    @classmethod
    def params(cls) -> dict:
        """Get the google credential file path.

        :return: [description]
        :rtype: [type]
        """

        if len(cls.ABSOLUTE_PATH_TO_CREDENTIAL_FILE) > 0:
            credentials = cls.ABSOLUTE_PATH_TO_CREDENTIAL_FILE
        else:

            abs_path_file = pathlib.Path(__file__).parent.absolute()

            credentials = os.path.join(abs_path_file, cls._RELATIVE_CREDENTIALS_DIR_PATH,
                                       cls._GOOGLE_APPLICATION_CREDENTIALS_FILENAME)
        return {
            "credentials": credentials
        }






from alternat.generation.config import Config
from alternat.generation.microsoft.config import Config as MicrosoftAnalyzerConfig
from alternat.generation.google.config import Config as GoogleAnalyzerConfig
from alternat.generation.opensource.config import Config as OpenAnalyzerConfig

from alternat.generation.microsoft.analyze import AnalyzeImage as MicrosoftAnalyzer
from alternat.generation.google.analyze import AnalyzeImage as GoogleAnalyzer
from alternat.generation.opensource.analyze import AnalyzeImage as OpenAnalyzer

from alternat.generation.rules.caption_handler import CaptionDataHandler
from alternat.generation.rules.ocr_handler import OCRDataHandler
from alternat.generation.rules.label_handler import LabelDataHandler


from .exceptions import InvalidGeneratorDriver, OutputDirPathNotGiven


class Drivers:
    """Driver name for alternat Library.
    """
    OPEN = "opensource"
    MICROSOFT = "azure"
    GOOGLE = "google"


class Generator:
    """Generator class to implement alternat Library.

    :raises InvalidGeneratorDriver: Driver Invalid
    :raises InvalidGeneratorDriver: Driver Invalid
    :raises InvalidGeneratorDriver: Driver Invalid
    :raises OutputDirPathNotGiven: Output director path is not given.
    :return: [description]
    :rtype: [type]
    """

    DEFAULT_DRIVER = Drivers.OPEN

    ALLOWED_DRIVERS = [
        Drivers.OPEN,
        Drivers.MICROSOFT,
        Drivers.GOOGLE
    ]

    def __init__(self, driver_name: str = None):
        """Initializes generator with driver to use.

        :param driver_name: Name of the driver], defaults to None (opensource)
        :type driver_name: str, optional
        :raises InvalidGeneratorDriver: Driver name is invalid or not implemented.
        """
        self.conf = Config()
        self.CURRENT_DRIVER = self.DEFAULT_DRIVER

        if driver_name is not None:
            if driver_name in self.ALLOWED_DRIVERS:
                self.CURRENT_DRIVER = driver_name
            else:
                raise InvalidGeneratorDriver(self.ALLOWED_DRIVERS)

        self._set_current_driver()

    def get_current_driver(self):
        """Get the current driver.

        :return: [description]
        :rtype: [type]
        """
        return self.CURRENT_DRIVER

    def get_config(self):
        """Get the generator level config in the form of JSON.

        :return: [description]
        :rtype: [type]
        """
        var = {}
        for v in Config.__dict__:
            if not callable(getattr(Config, v)):
                if not v.startswith("_") and not v.startswith("__"):
                    var[v] = Config.__dict__[v]
        return var

    def set_config(self, conf):
        """Sets the generator level configuration parameters passed via JSON.

        :param conf: Generator configuration parameters with values.
        :type conf: [type]
        """
        for attr in conf.keys():
            if attr in self.get_config().keys():
                setattr(Config, attr, conf[attr])

    def _set_current_driver(self):
        """Sets the current driver internally within the application.

        :raises InvalidGeneratorDriver: Driver name is invalid or not implemented.
        """
        if self.CURRENT_DRIVER == Drivers.OPEN:
            setattr(Config, Config.CURRENT_ANALYZER.__name__, OpenAnalyzer)
        elif self.CURRENT_DRIVER == Drivers.MICROSOFT:
            setattr(Config, Config.CURRENT_ANALYZER.__name__, MicrosoftAnalyzer)
        elif self.CURRENT_DRIVER == Drivers.GOOGLE:
            setattr(Config, Config.CURRENT_ANALYZER.__name__, GoogleAnalyzer)
        else:
            raise InvalidGeneratorDriver(self.ALLOWED_DRIVERS)

    def _get_current_driver_conf_cls(self):
        """Retreives the driver configuration class based on the currently driver

        :raises InvalidGeneratorDriver: Driver name is invalid or not implemented.
        :return: [description]
        :rtype: [type]
        """
        current_driver_cls = None
        if self.CURRENT_DRIVER == Drivers.OPEN:
            current_driver_cls = OpenAnalyzerConfig
        elif self.CURRENT_DRIVER == Drivers.MICROSOFT:
            current_driver_cls = MicrosoftAnalyzerConfig
        elif self.CURRENT_DRIVER == Drivers.GOOGLE:
            current_driver_cls = GoogleAnalyzerConfig
        else:
            raise InvalidGeneratorDriver(self.ALLOWED_DRIVERS)

        return current_driver_cls

    def get_driver_config(self):
        """Get the driver config in the form of JSON. Retreives public members [name: value] pair
        from the driver config class.

        :return: [description]
        :rtype: [type]
        """
        current_driver_conf_cls = self._get_current_driver_conf_cls()

        var = {}
        for v in current_driver_conf_cls.__dict__:
            if not callable(getattr(current_driver_conf_cls, v)):
                if not v.startswith("_") and not v.startswith("__"):
                    var[v] = current_driver_conf_cls.__dict__[v]
        return var

    def set_driver_config(self, conf: dict):
        """Sets the driver config parameters using the JSON passed. There is one-to-one 
        mapping between key in json and driver class public members.

        :param conf: Configuration JSON to set the driver configuration.
        :type conf: dict
        """
        current_driver_conf_cls = self._get_current_driver_conf_cls()

        for attr in conf.keys():
            if attr in self.get_driver_config().keys():
                setattr(current_driver_conf_cls, attr, conf[attr])

    def _process_image(self, image_path: str = None, output_dir_path: str = None, base64_image: str = None,
                       save: bool = True):
        """Process the image based on current driver and its configuration.

        :param image_path: Path to image on disk, defaults to None
        :type image_path: str, optional
        :param output_dir_path: Path to output directory where the results are stored, defaults to None
        :type output_dir_path: str, optional
        :param base64_image: Base64 image string, defaults to None
        :type base64_image: str, optional
        :param save: Whether to save the results in a file at specified location, defaults to True
        :type save: bool, optional
        :raises OutputDirPathNotGiven: Output directory path not given.
        :return: [description]
        :rtype: [type]
        """
        analyzer = Config.get_analyzer()

        data = analyzer.handle(image_path, base64_image)

        # TODO: Intermediate results should be saved separately
        # analyzer.save_result(data, output_dir_path)

        analyzer_conf = analyzer.get_conf()
        ocr_filter_threshold = analyzer_conf.get_ocr_height_width_to_image_ratio()
        caption_handler = CaptionDataHandler(data, analyzer_conf.get_caption_threshold())
        ocr_handler = OCRDataHandler(
            data, analyzer_conf.get_ocr_threshold(), ocr_filter_threshold
        )
        label_handler = LabelDataHandler(data, analyzer_conf.get_label_threshold())

        # Chain of responsibility
        caption_handler.set_next(ocr_handler).set_next(label_handler)
        output = caption_handler.handle()

        # the first handler in the chain also adds metadata information and saves the file to disk
        final_json = caption_handler.add_metadata(output)

        if save:
            if output_dir_path is not None:
                rules_json_filepath = caption_handler.save_result(final_json, output_dir_path)
            else:
                raise OutputDirPathNotGiven()

        return final_json

    def generate_alt_text_from_base64(self, base64_image: str):
        """Generates alt-text from base64 image string.

        :param base64_image: base64 image string
        :type base64_image: str
        :return: [description]
        :rtype: [type]
        """
        result_json = self._process_image(None, None, base64_image, False)
        return result_json

    def generate_alt_text_from_file(self, input_image_path: str, output_dir_path: str):
        """Generates alt-text from file on disk.

        :param input_image_path: Path to image to be processed.
        :type input_image_path: str
        :param output_dir_path: Path to directory where the results needs to be saved.
        :type output_dir_path: str
        :return: [description]
        :rtype: [type]
        """
        result_json = self._process_image(input_image_path, output_dir_path)
        return result_json






from alternat.generation.base.config import AnalyzerConf


class Config(AnalyzerConf):
    """Configuration class for Open source driver.

    :param AnalyzerConf: Base configuration class for driver.
    :type AnalyzerConf: [type]
    """

    # driver specific threshold for caption (used by rule engine)
    CAPTION_CONFIDENCE_THRESHOLD = 0.20

    # driver specific threshold for OCR (used by rule engine)
    OCR_CONFIDENCE_THRESHOLD = 0.30

    # driver specific threshold for label data (used by rule engine)
    LABEL_CONFIDENCE_THRESHOLD = 0.75

    # line height to image height ratio for OCR
    OCR_HEIGHT_RATIO_TO_IMAGE_THRESHOLD = 0.015





class SupportedActions:
    """Supported actions in generator driver. The driver implemented should support these actions.

    :return: [description]
    :rtype: [type]
    """
    OCR = "ocr"
    DESCRIBE = "caption"
    LABELS = "labels"

    _METADATA = "metadata"

    @classmethod
    def get_all(cls) -> list:
        """Get the list of all the actions in the generator driver.

        :return: [description]
        :rtype: list
        """
        return [cls.OCR, cls.DESCRIBE, cls.LABELS]

    @classmethod
    def get_metadata_key(cls):
        """Get the name of the metadata key.

        :return: [description]
        :rtype: [type]
        """
        return cls._METADATA



from .config import Config
from alternat.generation.base.analyzer import AnalyzeImageBase
import os, json, time
import requests, functools
from PIL import Image as PIL_Image
import easyocr
from alternat.generation.exceptions import InputImageNotAvailable


class AnalyzeImage(AnalyzeImageBase):
    """Opensource driver class.

    :param AnalyzeImageBase: Driver base class.
    :type AnalyzeImageBase: [type]
    """
    def __init__(self):
        super(AnalyzeImage, self).__init__()
        self.config = Config
        self.reader = easyocr.Reader(['en'])

    def modifyBoundingBoxData(self, bounding_box: list):
        """Transform bounding box data as per the convention. EasyOCR return bounding box info in the format
        [left, top, right, top, right, bottom, left, bottom] which is transformed to format
        [{x: left, y: top}, {x: right, y: top}, {x: right, y: bottom}, {x: left, y: bototm}].

        :param bounding_box: Bounding box data form EasyOCR.
        :type bounding_box: list
        :return: [description]
        :rtype: [type]
        """

        arr = bounding_box
        return [
            {"x": arr[0], "y": arr[1]},
            {"x": arr[2], "y": arr[3]},
            {"x": arr[4], "y": arr[5]},
            {"x": arr[6], "y": arr[7]}
        ]

    def ocr_analysis(self, image: PIL_Image):
        """Does OCR Analysis using EasyOCR.

        :param image: PIL Image object.
        :type image: PIL_Image
        """

        result = self.reader.readtext(self.pil_to_image_content(image))

        lines_data = []
        text = ""

        for line in result:
            confidence = line[2]
            bounding_box_data = line[0]
            line_text = line[1]
            bounding_box_arr = [float(bounding_box_data[0][0]), float(bounding_box_data[0][1]), float(bounding_box_data[1][0]),
                                float(bounding_box_data[1][1]), float(bounding_box_data[2][0]), float(bounding_box_data[2][1]),
                                float(bounding_box_data[3][0]), float(bounding_box_data[3][1])]

            lines_data.append({
                "confidence": float(round(confidence, 2)),
                "text": line_text,
                "boundingBox": self.modifyBoundingBoxData(bounding_box_arr)
            })

            text += line_text + "\n"

        final_ocr_data = {
            "text": text,
            "lines": lines_data
        }

        self.data[self.actions.OCR] = final_ocr_data

    # TODO: Add open source implementation for image captioning
    def describe_image(self, image: PIL_Image):
        """Describe image using open source solution. Not implemented right now.

        :param image: PIL Image object
        :type image: PIL_Image
        """

        self.data[self.actions.DESCRIBE] = {}

    # TODO: Add open source implementation for image labelling
    def extract_labels(self, image: PIL_Image):
        """Extract labels of image using open source solution. Not implemented righ now.

        :param image: PIL Image object.
        :type image: PIL_Image
        """

        self.data[self.actions.LABELS] = {}

    def handle(self, image_path: str = None, base64_image: str = None, actions: list = None) -> dict:
        """Entry point for the driver. Implements all the action and generates data for rule engine.

        :param image_path: Path to image on disk, defaults to None
        :type image_path: str, optional
        :param base64_image: Base64 image string, defaults to None
        :type base64_image: str, optional
        :param actions: list of actions to run, defaults to None (all actions execute)
        :type actions: list, optional
        :return: [description]
        :rtype: dict
        """

        try:
            im = self.extract_metadata(base64_image, image_path)
        except InputImageNotAvailable as e:
            print("ERROR: %s" % e)
            return self.data

        if actions is None:
            actions = self.actions.get_all()

        for action in actions:
            # if feature is supported
            if action in self.actions.get_all():

                if action == self.actions.OCR:
                    self.ocr_analysis(im)
                if action == self.actions.LABELS:
                    self.extract_labels(im)
                if action == self.actions.DESCRIBE:
                    self.describe_image(im)

        return self.data







from alternat.generation.base.action_data_handler import ActionDataHandler
from alternat.generation.config import Config as GenerationConfig


class CaptionDataHandler(ActionDataHandler):
    """Rule for processing caption data from driver.

    :param ActionDataHandler: Base class for rule.
    :type ActionDataHandler: [type]
    """
    def __init__(self, input_data: dict, confidence_threshold: float = None):
        """Initialize the handler with input data and confidence threshold (if available)

        :param input_data: Data from driver.
        :type input_data: dict
        :param confidence_threshold: Confidence threshold to filter captions with low threshold, defaults to None (Driver config default)
        :type confidence_threshold: float, optional
        """
        super(CaptionDataHandler, self).__init__(input_data)

        # CONFIDENCE THRESHOLD FILTER
        if confidence_threshold is None:
            self.CONFIDENCE_THRESHOLD = 0.20
        else:
            self.CONFIDENCE_THRESHOLD = confidence_threshold

        print("Caption Threshold : ", self.CONFIDENCE_THRESHOLD)

        self.PREFIX_TEXT = "Appears to be: "

    def has_data(self) -> bool:
        """Checks whethere caption data is avalible in the input data.

        :return: [description]
        :rtype: bool
        """

        if self.actions.DESCRIBE in self.input_data.keys():

            caption = self.input_data[self.actions.DESCRIBE]

            if len(caption) > 0:
                return True
            else:
                return False
        else:
            return False

    def apply(self, interim_result: dict) -> dict:
        """Process intermin result from previous rules in the chain and run caption rule.

        :param interim_result: Intermediate results from previous rules in the chain.
        :type interim_result: dict
        :return: [description]
        :rtype: dict
        """

        if self.has_data():
            caption_data = self.input_data[self.actions.DESCRIBE]

            if caption_data["confidence"] >= self.CONFIDENCE_THRESHOLD and len(caption_data["text"].strip()) > 0:

                caption = self.PREFIX_TEXT + caption_data["text"].strip() + ". "

                if GenerationConfig.DEBUG:
                    interim_result[self.actions.DESCRIBE] = {"caption": caption, "confidence": caption_data["confidence"]}
                else:
                    interim_result[self.actions.DESCRIBE] = caption

                interim_result[self.alt_recommendation_key] = caption
            else:
                interim_result[self.actions.DESCRIBE] = ""
                interim_result[self.alt_recommendation_key] = ""

        else:
            interim_result[self.actions.DESCRIBE] = ""
            interim_result[self.alt_recommendation_key] = ""

        return interim_result




import os
import sys
import requests
# If you are using a Jupyter notebook, uncomment the following line.
# %matplotlib inline
import matplotlib.pyplot as plt
from PIL import Image
from io import BytesIO

# Add your Computer Vision subscription key and endpoint to your environment variables.
if 'COMPUTER_VISION_SUBSCRIPTION_KEY' in os.environ:
    subscription_key = os.environ['COMPUTER_VISION_SUBSCRIPTION_KEY']
else:
    print("\nSet the COMPUTER_VISION_SUBSCRIPTION_KEY environment variable.\n**Restart your shell or IDE for changes to take effect.**")
    sys.exit()

if 'COMPUTER_VISION_ENDPOINT' in os.environ:
    endpoint = os.environ['COMPUTER_VISION_ENDPOINT']

analyze_url = endpoint + "vision/v3.1/analyze"

# Set image_path to the local path of an image that you want to analyze.
# Sample images are here, if needed:
# https://github.com/Azure-Samples/cognitive-services-sample-data-files/tree/master/ComputerVision/Images
dir_image_path = "/images"

for filename in os.listdir(dir_image_path):
    if filename.endswith(".jpg"):
        x = input(" filename " + filename + ". Continue ? (y/n) ")

        if x.lower() == "y":
            file_path = os.path.join(dir_image_path, filename)
            # Read the image into a byte array
            image_data = open(file_path, "rb").read()
            headers = {'Ocp-Apim-Subscription-Key': subscription_key,
                       'Content-Type': 'application/octet-stream'}
            params = {'visualFeatures': 'Categories,Description'}
            response = requests.post(
                analyze_url, headers=headers, params=params, data=image_data)
            response.raise_for_status()

            # The 'analysis' object contains various fields that describe the image. The most
            # relevant caption for the image is obtained from the 'description' property.
            analysis = response.json()
            print(analysis)
            image_caption = analysis["description"]["captions"][0]["text"].capitalize()

            # Display the image and overlay it with the caption.
            image = Image.open(BytesIO(image_data))
            plt.imshow(image)
            plt.axis("off")
            _ = plt.title(image_caption, size="x-large", y=-0.1)
            plt.show()
        else:
            print("Skipping")



class AnalyzerConf:
    """Base configuration class (driver specific) used by the rule engine.

    :return: [description]
    :rtype: [type]
    """
    # driver specific threshold for caption (used by rule engine)
    CAPTION_CONFIDENCE_THRESHOLD = 0.20

    # driver specific threshold for OCR (used by rule engine)
    OCR_CONFIDENCE_THRESHOLD = 0.75

    # driver specific threshold for label data (used by rule engine)
    LABEL_CONFIDENCE_THRESHOLD = 0.50

    # OCR height width ratio to image height width should be greater
    # than this value to pass through rule engine
    OCR_HEIGHT_RATIO_TO_IMAGE_THRESHOLD = 0.20

    @classmethod
    def get_ocr_height_width_to_image_ratio(cls):
        """Get the ratio for line height / image height used to filter out irrelevant OCR text.

        :return: [description]
        :rtype: [type]
        """
        return cls.OCR_HEIGHT_RATIO_TO_IMAGE_THRESHOLD

    @classmethod
    def get_ocr_threshold(cls):
        """Get the confidence threshold to filter OCR data.

        :return: [description]
        :rtype: [type]
        """
        return cls.OCR_CONFIDENCE_THRESHOLD

    @classmethod
    def get_label_threshold(cls):
        """Get the confidence threshold to filter label data.

        :return: [description]
        :rtype: [type]
        """
        return cls.LABEL_CONFIDENCE_THRESHOLD

    @classmethod
    def get_caption_threshold(cls):
        """Get the confidence threshold to filter caption data.

        :return: [description]
        :rtype: [type]
        """
        return cls.CAPTION_CONFIDENCE_THRESHOLD



import json, os


def save_json_to_disk(dir_path: str, result: dict, rel_path: str):
    """Saves JSON object in a JSON file on disk.

    :param dir_path: Path to directory where result needs to be saved.
    :type dir_path: str
    :param result: Data to be written in the JSON file.
    :type result: dict
    :param rel_path: Relative path to file.
    :type rel_path: str
    """

    if not os.path.exists(dir_path):
        os.makedirs(dir_path)

    with open(rel_path, "w") as f:
        json.dump(result, f)



from .actions import SupportedActions
from .config import AnalyzerConf
from alternat.generation.utilities import save_json_to_disk
from PIL import Image
from io import BytesIO
import base64, os
from alternat.generation.exceptions import InputImageNotAvailable


class AnalyzeImageBase:
    """Base class to implement generator driver.
    """
    def __init__(self):
        self.actions = SupportedActions()
        self.data = {key: "" for key in self.actions.get_all()}
        self.filename_postfix = "analyzer"
        self.config = AnalyzerConf

    @staticmethod
    def _base64_to_pil(base64_image):
        """Creates PIL Image object from base64 image

        :param base64_image: base64 image string.
        :type base64_image: [type]
        :return: [description]
        :rtype: [type]
        """
        im = Image.open(BytesIO(base64.b64decode(base64_image)))
        return im

    @staticmethod
    def _local_image_to_pil(abs_image_path: str):
        """Creates PIL Image object from image on disk.

        :param abs_image_path: Absolute path to image on disk.
        :type abs_image_path: str
        :return: [description]
        :rtype: [type]
        """
        im = Image.open(abs_image_path)
        return im

    @staticmethod
    def pil_to_image_content(im: Image, image_format: str = None):
        """Generate content of image from PIL Image object

        :param im: PIL Image object
        :type im: Image
        :param image_format: Image format (jpg / jpeg / png), defaults to None (JPEG)
        :type image_format: str, optional
        :return: [description]
        :rtype: [type]
        """
        buffered = BytesIO()

        if image_format is None:

            # convert to rgb format
            rgb_im = im.convert('RGB')
            img_format = "JPEG"
            rgb_im.save(buffered, format=img_format)
        else:
            img_format = image_format
            im.save(buffered, format=img_format)

        content = buffered.getvalue()
        return content

    def extract_metadata(self, base64_image: str = None, image_path: str = None):
        """Process the image to add metadata

        :param base64_image: base64 image string, defaults to None
        :type base64_image: str, optional
        :param image_path: path to image on disk, defaults to None
        :type image_path: str, optional
        :raises InputImageNotAvailable: Error when both base64_image and image_path is not there - one should be present.
        :return: [description]
        :rtype: [type]
        """
        if base64_image is not None:
            im = self._base64_to_pil(base64_image)
            self.add_metadata_base64(im, base64_image)
            return im
        elif image_path is not None:
            abs_image_path = os.path.abspath(image_path)
            im = self._local_image_to_pil(abs_image_path)
            self.add_metadata_local_file(im, abs_image_path)
            return im
        else:
            raise InputImageNotAvailable()

    def add_metadata_local_file(self, im: Image, abs_image_path: str):
        """Add metadata when the image is available locally on disk

        :param im: PIL Image object.
        :type im: Image
        :param abs_image_path: Absolute path to image on disk.
        :type abs_image_path: str
        """
        width, height = im.size
        image_type = abs_image_path.split(".")[-1]

        self.data[self.actions.get_metadata_key()] = {
            "filename": abs_image_path.split("/")[-1],
            "filepath": abs_image_path,
            "storageType": "local",
            "width": width,
            "height": height,
            "type": image_type
        }

    def add_metadata_base64(self, im: Image, base64_image: str):
        """Adds metadata when the image is base 64 image.

        :param im: PIL Image object.
        :type im: Image
        :param base64_image: base64 image string.
        :type base64_image: str
        """
        width, height = im.size
        self.data[self.actions.get_metadata_key()] = {
            "filename": "",
            "filepath": "",
            "storageType": "base64",
            "width": width,
            "height": height,
            "type": ""
        }

    def get_conf(self):
        """Get the configuration class for driver.

        :return: Config class of driver
        :rtype: [type]
        """
        return self.config

    def handle(self, image_path: str = None, base64_image: str = None, actions: list = None) -> dict:
        """Entry point for the driver. Implements all the action and generates data for rule engine.

        :param image_path: Path to image on disk, defaults to None
        :type image_path: str, optional
        :param base64_image: Base64 image string, defaults to None
        :type base64_image: str, optional
        :param actions: list of actions to run, defaults to None (all actions execute)
        :type actions: list, optional
        :return: [description]
        :rtype: dict
        """
        return self.data

    def is_clean(self, abs_image_path: str) -> bool:
        """Check if the image is clean.

        :param abs_image_path: Absolute path to image on disk.
        :type abs_image_path: str
        :return: [description]
        :rtype: bool
        """
        return True

    def save_result(self, result: dict, output_dir_path: str):
        """Saves result to output folder in disk.

        :param result: Result JSON genearted by driver.
        :type result: dict
        :param output_dir_path: Path to directory where the results should be saved.
        :type output_dir_path: str
        :return: [description]
        :rtype: [type]
        """
        dir_path = output_dir_path
        filename = result[self.actions.get_metadata_key()]["filename"]
        name = filename.rsplit(".", 1)[0] + "_" + self.filename_postfix
        filepath = dir_path + "/" + name + ".json"
        save_json_to_disk(dir_path, result, filepath)
        return filepath






from __future__ import annotations
from .actions import SupportedActions
from alternat.generation.utilities import save_json_to_disk
import os

class Handler(object):
    """Implements chain of responsibilty for rule engine handlers.

    :param object: [description]
    :type object: [type]
    """
    def __init__(self):
        self._next_handler = None

    def set_next(self, handler: Handler) -> Handler:
        """Sets the next handler in the chain.

        :param handler: Rule engine handler to be implemented in as the next item in chain.
        :type handler: Handler
        :return: [description]
        :rtype: Handler
        """

        self._next_handler = handler
        return handler

    def handle(self, data: dict):
        """
        Handles the request
        :return:
        """
        pass


class ActionDataHandler(Handler):
    """Base class for rule engine handlers.

    :param Handler: [description]
    :type Handler: [type]
    """
    def __init__(self, input_data: dict):
        """Initializes input data for the rule.

        :param input_data: [description]
        :type input_data: dict
        """
        super(ActionDataHandler, self).__init__()
        self.input_data = input_data
        self.actions = SupportedActions()

        self.alt_recommendation_key = "alt"
        self.filename_postfix = "rules"

    def handle(self, interim_result: dict = None):
        """Calls the next handler after processing interim results.

        :param interim_result: Intermediate results from previous rules in the chain, defaults to None.
        :type interim_result: dict, optional
        :return: [description]
        :rtype: [type]
        """

        if interim_result is None:
            result = self.apply({})
        else:
            result = self.apply(interim_result)

        if self._next_handler is not None:
            return self._next_handler.handle(result)
        else:
            return result

    def add_metadata(self, result: dict):
        """Adds metadata to the final result.

        :param result: [description]
        :type result: dict
        :return: [description]
        :rtype: [type]
        """

        if self.actions.get_metadata_key() in self.input_data.keys():
            result[self.actions.get_metadata_key()] = self.input_data[self.actions.get_metadata_key()]
        else:
            result[self.actions.get_metadata_key()] = ''
        return result

    def save_result(self, result: dict, output_dir):
        """Saves result on disk.

        :param result: JSON data to be saved.
        :type result: dict
        :param output_dir: Directory where the data needs to be saved.
        :type output_dir: [type]
        :return: [description]
        :rtype: [type]
        """
        dir_path = output_dir
        filename = result[self.actions.get_metadata_key()]["filename"]
        name = filename.rsplit(".", 1)[0] + "_" + self.filename_postfix
        filepath = os.path.join( dir_path , name + ".json")
        save_json_to_disk(dir_path, result, filepath)
        return filepath

    def has_data(self) -> bool:
        """Checks if handler has relevant data.

        :return: [description]
        :rtype: bool
        """
        return False

    def apply(self, interim_result: dict) -> dict:
        """Processes the interim result based on the rule.

        :param interim_result: Intermediate results from previous rules in the chain.
        :type interim_result: dict
        :return: [description]
        :rtype: dict
        """

        return self.input_data



from alternat.generation.microsoft.analyze import AnalyzeImage as ImageAnalyzerMicrosoft
from alternat.generation.google.analyze import AnalyzeImage as ImageAnalyzerGoogle
from alternat.generation.opensource.analyze import AnalyzeImage as ImageAnalyzerOpen


from alternat.generation.base.analyzer import AnalyzeImageBase


class Config:
    """Configuration class for generation.

    :return: [description]
    :rtype: [type]
    """

    # Change this to "ImageAnalyzerGoogle" if you want to use Google Analyzer
    # Change this to "ImageAnalyzerMicrosoft" if you want to use Azure Analyzer
    CURRENT_ANALYZER = ImageAnalyzerMicrosoft

    # when True, the output will also save confidence value for OCR with line height data
    # This is useful for optimizing OCR thresholds.
    DEBUG = False

    # this enables ocr clustering
    ENABLE_OCR_CLUSTERING = True

    @classmethod
    def get_analyzer(cls) -> AnalyzeImageBase:
        """Get the current analyzer.

        :return: [description]
        :rtype: AnalyzeImageBase
        """
        analyzer = getattr(cls, cls.CURRENT_ANALYZER.__name__)
        return analyzer()



class AlternatException(Exception):
    """Alternat Exception base class.

    :param Exception: [description]
    :type Exception: [type]
    """

    # Error text to be displayed
    ERROR_TEXT = ""

    def __init__(self):
        self.message = self.ERROR_TEXT
        super(AlternatException, self).__init__(self.message)


class InputImageNotAvailable(AlternatException):
    """Input image not availble exception.

    :param AlternatException: [description]
    :type AlternatException: [type]
    """

    # Error text to be displayed
    ERROR_TEXT = "Input image not found: Either base64 image or local file path is required."


class OutputDirPathNotGiven(AlternatException):
    """Output directory path not specified.

    :param AlternatException: [description]
    :type AlternatException: [type]
    """

    # Error text to be displayed
    ERROR_TEXT = "Output directory not provided to save the result."


class InvalidConfigFile(AlternatException):
    """Invalid JSON for configuration.

    :param AlternatException: [description]
    :type AlternatException: [type]
    """

    # Error text to be displayed
    ERROR_TEXT = "Need a JSON file to parse configuration. " \
                 "See sample config of Generation drivers or refer documentation for more info."


class InvalidGeneratorDriver(Exception):
    """Driver is invalid or not implemented.

    :param Exception: [description]
    :type Exception: [type]
    """

    # Error text to be displayed
    ERROR_TEXT = "Driver name is invalid !"

    def __init__(self, allowed_drivers: list):
        self.allowed_drivers = allowed_drivers
        self.message = self.ERROR_TEXT + " Allowed drivers are : " + ", ".join(self.allowed_drivers) + "."
        super(InvalidGeneratorDriver, self).__init__(self.message)


class OCRClusteringException(AlternatException):
    """Cannot process image via alternat clustering algorithm.

    :param AlternatException: [description]
    :type AlternatException: [type]
    """

    # Error text to be displayed
    ERROR_TEXT = "Clustering of OCR data is not possible."




import os
import sys
import requests
# If you are using a Jupyter notebook, uncomment the following line.
# %matplotlib inline
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
from PIL import Image
from io import BytesIO

# Add your Computer Vision subscription key and endpoint to your environment variables.
if 'COMPUTER_VISION_SUBSCRIPTION_KEY' in os.environ:
    subscription_key = os.environ['COMPUTER_VISION_SUBSCRIPTION_KEY']
else:
    print("\nSet the COMPUTER_VISION_SUBSCRIPTION_KEY environment variable.\n**Restart your shell or IDE for changes to take effect.**")
    sys.exit()

if 'COMPUTER_VISION_ENDPOINT' in os.environ:
    endpoint = os.environ['COMPUTER_VISION_ENDPOINT']

analyze_url = endpoint + "vision/v3.1/ocr"

# Set image_path to the local path of an image that you want to analyze.
# Sample images are here, if needed:
# https://github.com/Azure-Samples/cognitive-services-sample-data-files/tree/master/ComputerVision/Images
dir_image_path = "/images_with_text"

for filename in os.listdir(dir_image_path):
    if filename.endswith(".jpg") or filename.endswith(".jpeg") or filename.endswith(".svg") or filename.endswith(".png"):
        x = input(" filename " + filename + ". Continue ? (y/n) ")

        if x.lower() == "y":
            file_path = os.path.join(dir_image_path, filename)
            # Read the image into a byte array
            image_data = open(file_path, "rb").read()
            headers = {'Ocp-Apim-Subscription-Key': subscription_key,
                       'Content-Type': 'application/octet-stream'}
            params = {'language': 'en', 'detectOrientation': 'true'}

            try:
                response = requests.post(
                    analyze_url, headers=headers, params=params, data=image_data)
                response.raise_for_status()
            except requests.exceptions.HTTPError:
                print(response.content)
                continue


            # The 'analysis' object contains various fields that describe the image. The most
            # relevant caption for the image is obtained from the 'description' property.
            analysis = response.json()
            print(analysis)
            # Extract the word bounding boxes and text.
            line_infos = [region["lines"] for region in analysis["regions"]]
            word_infos = []
            for line in line_infos:
                for word_metadata in line:
                    for word_info in word_metadata["words"]:
                        word_infos.append(word_info)
            word_infos

            # Display the image and overlay it with the extracted text.
            plt.figure(figsize=(5, 5))
            image = Image.open(BytesIO(image_data))
            print(" type of image ", type(image))
            ax = plt.imshow(image)
            for word in word_infos:
                bbox = [int(num) for num in word["boundingBox"].split(",")]
                text = word["text"]
                origin = (bbox[0], bbox[1])
                patch = Rectangle(origin, bbox[2], bbox[3],
                                  fill=False, linewidth=2, color='y')
                ax.axes.add_patch(patch)
                plt.text(origin[0], origin[1], text, fontsize=20, weight="bold", va="top")
            plt.show()
            plt.axis("off")
        else:
            print("Skipping")



from .config import Config
from alternat.generation.base.analyzer import AnalyzeImageBase
import os, json, time
import time
import requests, functools
from PIL import Image as PIL_Image
import json
from alternat.generation.exceptions import InputImageNotAvailable


class AnalyzeImage(AnalyzeImageBase):
    """Azure / Microsoft Analyzer driver class.

    :param AnalyzeImageBase: Driver base class.
    :type AnalyzeImageBase: [type]
    """
    def __init__(self):
        super(AnalyzeImage, self).__init__()
        self.config = Config

        self.params = self.config.params()

        self.describe_endpoint = self.params["endpoint"] + "vision/v3.1/analyze"
        self.ocr_endpoint = self.params["endpoint"] + "vision/v3.1/read/analyze"

    def modifyBoundingBoxData(self, bounding_box: list):
        """Transform bounding box data as per the convention. Azure API return bounding box info in the format
        [left, top, right, top, right, bottom, left, bottom] which is transformed to format
        [{x: left, y: top}, {x: right, y: top}, {x: right, y: bottom}, {x: left, y: bototm}].

        :param bounding_box: Bounding box data form Azure API.
        :type bounding_box: list
        :return: [description]
        :rtype: [type]
        """

        arr = bounding_box
        return [
            {"x": arr[0], "y": arr[1]},
            {"x": arr[2], "y": arr[3]},
            {"x": arr[4], "y": arr[5]},
            {"x": arr[6], "y": arr[7]}
        ]

    def ocr_analysis(self, image: PIL_Image):
        """Does OCR Analysis using Azure Vision API.

        :param image: PIL Image object.
        :type image: PIL_Image
        """

        headers = {'Ocp-Apim-Subscription-Key': self.params["subscription"],
                   'Content-Type': 'application/octet-stream'}
        params = {'language': 'en', 'detectOrientation': 'true'}

        try:
            response = requests.post(
                self.ocr_endpoint, headers=headers, params=params, data=self.pil_to_image_content(image))
            response.raise_for_status()
        except requests.exceptions.HTTPError:
            exception = json.loads(response.content)

            if exception["error"]["code"] == "429":

                if Config.AZURE_RATE_LIMIT_ON:
                    print("\n\nRate limited by Azure. The next call will execute after %d sec \n\n" %
                          Config.AZURE_RATE_LIMIT_TIME_IN_SEC)

                    time.sleep(Config.AZURE_RATE_LIMIT_TIME_IN_SEC)
                    self.ocr_analysis(image)
                    return

                else:
                    print("\n\nRate limited by Azure. You can enable rate limiting by "
                          "setting AZURE_RATE_LIMIT_ON in config.py\n\n")
                    return

            else:
                print("Error in sending request: ", response.content)
                return

        # Holds the URI used to retrieve the recognized text.
        operation_url = response.headers["Operation-Location"]

        # The recognized text isn't immediately available, so poll to wait for completion.
        analysis = {}
        poll = True
        while poll:
            response_final = requests.get(
                response.headers["Operation-Location"], headers=headers)
            analysis = response_final.json()

            # print(json.dumps(analysis, indent=4))

            time.sleep(1)
            if "analyzeResult" in analysis:
                poll = False
            if "status" in analysis and analysis['status'] == 'failed':
                poll = False

        lines_data = []
        text = ""

        if "analyzeResult" in analysis:
            # Extract the recognized text, with bounding boxes.
            for line in analysis["analyzeResult"]["readResults"][0]["lines"]:

                words = line["words"]
                total_words_confidence = functools.reduce(lambda a, b: a + b,
                                                          [word["confidence"] for word in words])
                average_confidence = round(total_words_confidence / len(words), 2)
                lines_data.append({
                    "text": line["text"],
                    "confidence": average_confidence,
                    "boundingBox": self.modifyBoundingBoxData(line["boundingBox"])
                })

                text += line["text"] + "\n"

        final_ocr_data = {
            "text": text,
            "lines": lines_data
        }

        self.data[self.actions.OCR] = final_ocr_data

    def describe_image(self, image: PIL_Image):
        """Describe image using Azure Vision API.

        :param image: PIL Image object
        :type image: PIL_Image
        """

        headers = {'Ocp-Apim-Subscription-Key': self.params["subscription"],
                   'Content-Type': 'application/octet-stream'}
        params = {'visualFeatures': 'Categories,Description'}

        try:
            response = requests.post(
                self.describe_endpoint, headers=headers, params=params, data=self.pil_to_image_content(image))

            response.raise_for_status()
        except requests.exceptions.HTTPError:
            exception = json.loads(response.content)

            if exception["error"]["code"] == "429":

                if Config.AZURE_RATE_LIMIT_ON:
                    print("\n\nRate limited by Azure. The next call will execute after %d sec \n\n" %
                          Config.AZURE_RATE_LIMIT_TIME_IN_SEC)

                    time.sleep(Config.AZURE_RATE_LIMIT_TIME_IN_SEC)
                    self.describe_image(image)
                    return

                else:
                    print("\n\nRate limited by Azure. You can enable rate limiting by "
                          "setting AZURE_RATE_LIMIT_ON in config.py\n\n")
                    return

            else:
                print("Error in sending request: ", response.content)
                return

        analysis = response.json()
        # print(" DESCRIBE : ", analysis)
        caption = analysis["description"]["captions"][0]["text"]
        confidence = analysis["description"]["captions"][0]["confidence"]

        self.data[self.actions.DESCRIBE] = {"text": caption, "confidence": confidence}

    def extract_labels(self, image: PIL_Image):
        """Extract labels of image using Azure Vision API.

        :param image: PIL Image object.
        :type image: PIL_Image
        """
        headers = {'Ocp-Apim-Subscription-Key': self.params["subscription"],
                   'Content-Type': 'application/octet-stream'}
        params = {'visualFeatures': 'Tags'}

        try:
            response = requests.post(
                self.describe_endpoint, headers=headers, params=params, data=self.pil_to_image_content(image))

            response.raise_for_status()
        except requests.exceptions.HTTPError:
            exception = json.loads(response.content)

            if exception["error"]["code"] == "429":

                if Config.AZURE_RATE_LIMIT_ON:
                    print("\n\nRate limited by Azure. The next call will execute after %d sec \n\n" %
                          Config.AZURE_RATE_LIMIT_TIME_IN_SEC)

                    time.sleep(Config.AZURE_RATE_LIMIT_TIME_IN_SEC)
                    self.extract_labels(image)
                    return

                else:
                    print("\n\nRate limited by Azure. You can enable rate limiting by "
                          "setting AZURE_RATE_LIMIT_ON in config.py\n\n")
                    return

            else:
                print("Error in sending request: ", response.content)
                return

        analysis = response.json()
        labels = analysis["tags"]

        labels_data = []

        for label in labels:
            labels_data.append({
                "description": label["name"],
                "confidence": label["confidence"]
            })

        self.data[self.actions.LABELS] = labels_data

    def resize_image(self, image: PIL_Image):
        """Resize image (maintaining aspect ratio) if width / height > 5000 pixels (API constrain from Azure)

        :param image: [description]
        :type image: PIL_Image
        """
        size = Config._MAX_IMAGE_SIZE_IN_PIXEL, Config._MAX_IMAGE_SIZE_IN_PIXEL
        image.thumbnail(size, PIL_Image.ANTIALIAS)
        # im.save(abs_image_path)

    def is_clean(self, image: PIL_Image) -> bool:
        """Check if the image has proper resolution, and is clean.

        :param image:PIL Image object.
        :type image: PIL_Image
        :return: [description]
        :rtype: bool
        """
        width, height = image.size

        if width < Config._MIN_IMAGE_SIZE_IN_PIXEL or height < Config._MIN_IMAGE_SIZE_IN_PIXEL:
            return False

        if width > Config._MAX_IMAGE_SIZE_IN_PIXEL or height > Config._MAX_IMAGE_SIZE_IN_PIXEL:
            self.resize_image(image)

        return True

    def handle(self, image_path: str = None, base64_image: str = None, actions: list = None) -> dict:
        """Entry point for the driver. Implements all the action and generates data for rule engine.

        :param image_path: Path to image on disk, defaults to None
        :type image_path: str, optional
        :param base64_image: Base64 image string, defaults to None
        :type base64_image: str, optional
        :param actions: list of actions to run, defaults to None (all actions execute)
        :type actions: list, optional
        :return: [description]
        :rtype: dict
        """

        try:
            im = self.extract_metadata(base64_image, image_path)
        except InputImageNotAvailable as e:
            print("ERROR: %s" % e)
            return self.data

        if self.is_clean(im):
            if actions is None:
                actions = self.actions.get_all()

            for action in actions:
                # if feature is supported
                if action in self.actions.get_all():

                    if action == self.actions.OCR:
                        self.ocr_analysis(im)
                    if action == self.actions.LABELS:
                        self.extract_labels(im)
                    if action == self.actions.DESCRIBE:
                        self.describe_image(im)
        else:
            print("Image dimensions outside of bound Min: %d X %d | Max : %d X %d"
                  % (Config._MIN_IMAGE_SIZE_IN_PIXEL, Config._MIN_IMAGE_SIZE_IN_PIXEL,
                     Config._MAX_IMAGE_SIZE_IN_PIXEL, Config._MAX_IMAGE_SIZE_IN_PIXEL))
            print("Skipping this image")

        return self.data







#!/usr/bin/env python
# -*- coding: utf-8 -*-
import re
import os
import sys
#import setuptools
import subprocess
from setuptools import setup
from setuptools import find_packages
from setuptools.command.develop import develop
from setuptools.command.install import install
from setuptools.command.egg_info import egg_info


with open("Readme.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()

# helper functions to make it easier to list dependencies not as a python list, but vertically w/ optional built-in comments to why a certain version of the dependency is listed
def cleanup(x):
    return re.sub(r" *#.*", "", x.strip())  # comments


def to_list(buffer):
    return list(filter(None, map(cleanup, buffer.splitlines())))


# normal dependencies ###
#
# these get resolved and installed via either of these two:
#
#   pip install alternat
#   pip install -e .
#
# IMPORTANT: when updating these, please make sure to sync conda/meta.yaml
dep_groups = {
    "core": to_list(
        """
        pillow
        google-cloud-vision==1.0.0
        tldextract
        easyocr
        pyyaml
        treelib
        uvicorn
        fastapi==0.62.0
        typer==0.3.2
"""
    )
}

__version__ = None # Explicitly set version.
# TODO use os.path instead of front slash
exec(open('alternat/version.py').read()) # loads __version__

requirements = [y for x in dep_groups.values() for y in x]
setup_requirements = to_list(
    """
    pytest-runner
    setuptools>=36.2
"""
)


# test dependencies ###
test_requirements = to_list(
    """
    pytest
"""
)

#mkdir my-project
#cd my-project
#npm install bitcoinjs-lib

setup(
    name="alternat",
    version=__version__,
    author="keplerlab",
    author_email="keplerwaasi@gmail.com",
    description="Alternat is a tool that automates alt text generation.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/keplerlab/alternat.git",
    packages=find_packages(),
    install_requires=requirements,
    setup_requires=setup_requirements,
    tests_require=test_requirements,
    python_requires=">=3.6",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    project_urls={
        "Documentation": "https://alternat.readthedocs.io",
        "Source": "https://github.com/keplerlab/alternat",
        "Tracker": "https://github.com/keplerlab/alternat/issues",
    },
    include_package_data=True,
    zip_safe=False,
)



from alternat.generation.base.action_data_handler import ActionDataHandler
from alternat.generation.config import Config as GenerationConfig
from .ocr_cluster import ClusterTree


class OCRDataHandler(ActionDataHandler):
    """Rule for processing OCR data from driver.

    :param ActionDataHandler: Base class for rule.
    :type ActionDataHandler: [type]
    """
    def __init__(self, input_data, confidence_threshold: float = None,
                 ocr_filter_threshold: float = None):
        """Initialize the handler with input data and confidence threshold (if available)

        :param input_data: Data from driver.
        :type input_data: [type]
        :param confidence_threshold: Confidence threshold to filter OCR with low threshold, defaults to None (Driver config default)
        :type confidence_threshold: float, optional
        :param ocr_filter_threshold: Confidence threshold to filter OCR data based on line height ratio to image height.
        :type ocr_filter_threshold: float, optional
        """
        super(OCRDataHandler, self).__init__(input_data)

        # CONFIDENCE THRESHOLD FOR OCR
        if confidence_threshold is None:
            self.CONFIDENCE_THRESHOLD = 0.75
        else:
            self.CONFIDENCE_THRESHOLD = confidence_threshold

        print("OCR Threshold : ", self.CONFIDENCE_THRESHOLD)

        self.OCR_HEIGHT_RATIO_TO_IMAGE_THRESHOLD = ocr_filter_threshold

        # AREA OCCUPIED THRESHOLD
        self.AREA_OCCUPIED_BY_TEXT_THRESHOLD = 0.30

        # Numbers of ocr data to show
        self.NUM_OF_OCR = 3

        self.PREFIX_TEXT = "Appears to contain text: "

        self.ocr_cluster_cls = ClusterTree()

    def has_data(self) -> bool:
        """Checks whethere OCR data is avalible in the input data.

        :return: [description]
        :rtype: bool
        """

        if self.actions.OCR in self.input_data.keys():
            ocr_data = self.input_data[self.actions.OCR]

            if len(ocr_data) > 0:
                return True
            else:
                return False
        else:
            return False

    def process_ocr(self) -> dict:
        """Process the OCR data from the driver and filter it on the basis of
        line confidence threshold value and the ratio of line height to image height. 
        Based on the configuration also invokes alternat clustering implementation (default to True)

        :return: [description]
        :rtype: dict
        """

        ocr_data = self.input_data[self.actions.OCR]
        lines_data = ocr_data["lines"]
        image_metadata = self.input_data[self.actions.get_metadata_key()]
        image_height = image_metadata["height"]
        image_width = image_metadata["width"]

        filtered_lines = list(filter(lambda l: l["confidence"] >= self.CONFIDENCE_THRESHOLD, lines_data))

        filtered_text = ""

        # only for debug
        debug_line_data = []

        total_line_area = 0

        final_lines_data = []

        for line in filtered_lines:
            line_width = abs(line["boundingBox"][1]["x"] - line["boundingBox"][0]["x"])
            line_height = abs(line["boundingBox"][3]["y"] - line["boundingBox"][0]["y"])
            line_area = line_width * line_height

            # print(" line height, image height ", line_height, image_height)

            if ((line_height / image_height) > self.OCR_HEIGHT_RATIO_TO_IMAGE_THRESHOLD):
                filtered_text += line["text"] + "\n"
                total_line_area += line_area

                final_lines_data.append(line)

            #debug array data has everything
            debug_line_data.append({
                "confidence": line["confidence"],
                "line_area": line_area,
                "text": line["text"],
                "height_ratio": round(line_height / image_height, 5),
                "height":line_height
            })

        image_area = image_height * image_width

        occupied_area_by_lines = total_line_area / image_area
        # print("Occupied Area :", occupied_area_by_lines)

        # use the default final text
        final_text = filtered_text.replace("\n", " ")

        if GenerationConfig.ENABLE_OCR_CLUSTERING:
            ocr_cluster_text = self.ocr_cluster_cls.generate_data(final_lines_data)

            if len(ocr_cluster_text) > 0:
                final_text = ocr_cluster_text

        if GenerationConfig.DEBUG:
            return {
                "text": final_text,
                "is_area_above_threshold": occupied_area_by_lines > self.AREA_OCCUPIED_BY_TEXT_THRESHOLD,
                "data": debug_line_data
            }

        else:
            return {
                "text": final_text,
                "is_area_above_threshold": occupied_area_by_lines > self.AREA_OCCUPIED_BY_TEXT_THRESHOLD
            }

    def apply(self, interim_result: dict) -> dict:
        """Process intermin result from previous rules in the chain and run OCR rule.

        :param interim_result: Intermediate results from previous rules in the chain.
        :type interim_result: dict
        :return: [description]
        :rtype: dict
        """

        if self.has_data():
            processed_ocr_data = self.process_ocr()

            if processed_ocr_data["is_area_above_threshold"]:
                alt_recommendation = self.PREFIX_TEXT + processed_ocr_data["text"] + ". " + \
                                     interim_result[self.alt_recommendation_key]
                interim_result[self.alt_recommendation_key] = alt_recommendation

                if GenerationConfig.DEBUG:
                    interim_result[self.actions.OCR] = processed_ocr_data["data"]
                else:
                    interim_result[self.actions.OCR] = self.PREFIX_TEXT + processed_ocr_data["text"]
            else:
                if len(processed_ocr_data["text"]) > 0:
                    alt_recommendation = interim_result[self.alt_recommendation_key] + self.PREFIX_TEXT + \
                                         processed_ocr_data["text"] + ". "

                    if GenerationConfig.DEBUG:
                        interim_result[self.actions.OCR] = processed_ocr_data["data"]
                    else:
                        interim_result[self.actions.OCR] = self.PREFIX_TEXT + processed_ocr_data["text"]
                else:
                    alt_recommendation = interim_result[self.alt_recommendation_key]
                    interim_result[self.actions.OCR] = ""

                interim_result[self.alt_recommendation_key] = alt_recommendation


        else:
            interim_result[self.actions.OCR] = ""

        return interim_result



from alternat.generation.base.analyzer import AnalyzeImageBase
from google.cloud import vision
import io, os
from .config import Config
from PIL import Image as PIL_IMAGE
from alternat.generation.exceptions import InputImageNotAvailable


class AnalyzeImage(AnalyzeImageBase):
    """Google Analyzer driver class.

    :param AnalyzeImageBase: Driver base class.
    :type AnalyzeImageBase: [type]
    """
    def __init__(self):
        super(AnalyzeImage, self).__init__()
        self.config = Config

        self.params = self.config.params()

        self.set_environment_variables()

    def set_environment_variables(self):
        """Sets environment variable GOOGLE_APPLICATION_CREDENTIALS based on config.
        """
        os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = self.params["credentials"]

    def describe_image(self, image: PIL_IMAGE):
        """Describe image (used for captioning) - Not availble in Google Computer Vision API

        :param image: [description]
        :type image: PIL_IMAGE
        """
        self.data[self.actions.DESCRIBE] = {"text": '', "confidence": 1}

    def ocr_analysis(self, image: PIL_IMAGE):
        """Does OCR analysis using Google Computer Vision API. Also runs the alternat clustering rule
        if app is configured for it.

        :param image: PIL Image object.
        :type image: PIL_IMAGE
        """

        client = vision.ImageAnnotatorClient()

        # with io.open(abs_image_path, 'rb') as image_file:
        #     content = image_file.read()

        image = vision.types.Image(content=self.pil_to_image_content(image))

        # response = client.annotate_image({
        #     "image": image,
        #     "features": [{"type": vision.enums.Feature.Type.TEXT_DETECTION}]
        # })

        response = client.document_text_detection(image=image)

        full_ocr_data = response.full_text_annotation

        final_ocr_data = {
            "text": full_ocr_data.text,
            "lines": []
        }

        lines_data = []

        pages = full_ocr_data.pages

        for page in pages:
            blocks = page.blocks

            # google doesnt not gives line level information so the blocks here become the lines
            for block in blocks:
                paragraphs = block.paragraphs
                block_text = ""
                for paragraph in paragraphs:
                    for word in paragraph.words:
                        symbols = word.symbols
                        for character in symbols:
                            block_text += character.text

                        # add a space between words
                        block_text += " "

                # a block of word has finished
                block_text += "."

                lines_data.append({
                    "confidence": round(block.confidence, 2),
                    "text": block_text,
                    "boundingBox": [{"x": coord.x, "y": coord.y} for coord in block.bounding_box.vertices]
                })

        final_ocr_data["lines"] = lines_data

        # self.data[self.actions.OCR] = lines_data
        self.data[self.actions.OCR] = final_ocr_data

    def extract_labels(self, image: PIL_IMAGE):
        """Extract labels of image using Google Computer Vision API.

        :param image: PIL Image object.
        :type image: PIL_IMAGE
        :raises Exception: Google Cloud specific error messages based on request.
        """

        client = vision.ImageAnnotatorClient()

        # with io.open(abs_image_path, 'rb') as image_file:
        #     content = image_file.read()

        image = vision.types.Image(content=self.pil_to_image_content(image))

        response = client.label_detection(image=image)
        labels = response.label_annotations
        if response.error.message:
            raise Exception(
                '{}\nFor more info on error messages, check: '
                'https://cloud.google.com/apis/design/errors'.format(
                    response.error.message))

        label_data = []
        for label in labels:
            label_data.append({
                "description": label.description,
                "confidence": label.score
            })

        self.data[self.actions.LABELS] = label_data

    def handle(self, image_path: str = None, base64_image: str = None, actions: list = None) -> dict:
        """Entry point for the driver. Implements all the action and generates data for rule engine.

        :param image_path: Path to image on disk, defaults to None
        :type image_path: str, optional
        :param base64_image: Base64 image string, defaults to None
        :type base64_image: str, optional
        :param actions: list of actions to run, defaults to None (all actions execute)
        :type actions: list, optional
        :return: [description]
        :rtype: dict
        """
        try:
            im = self.extract_metadata(base64_image, image_path)
        except InputImageNotAvailable as e:
            print("ERROR: %s" % e)
            return self.data

        if actions is None:
            actions = self.actions.get_all()

        for action in actions:
            # if feature is supported
            if action in self.actions.get_all():

                if action == self.actions.OCR:
                    self.ocr_analysis(im)
                if action == self.actions.LABELS:
                    self.extract_labels(im)
                if action == self.actions.DESCRIBE:
                    self.describe_image(im)

        return self.data



from alternat.generation.base.action_data_handler import ActionDataHandler
from alternat.generation.config import Config as GenerationConfig


class LabelDataHandler(ActionDataHandler):
    """Rule for processing label data from driver.

    :param ActionDataHandler: Base class for rule.
    :type ActionDataHandler: [type]
    """
    def __init__(self, input_data, confidence_threshold: float = None):
        """Initialize the handler with input data and confidence threshold (if available)

        :param input_data: Data from driver.
        :type input_data: [type]
        :param confidence_threshold: Confidence threshold to filter labels with low threshold, defaults to None (Driver config default)
        :type confidence_threshold: float, optional
        """
        super(LabelDataHandler, self).__init__(input_data)

        # CONFIDENCE THRESHOLD FOR LABELS
        if confidence_threshold is None:
            self.CONFIDENCE_THRESHOLD = 0.50
        else:
            self.CONFIDENCE_THRESHOLD = confidence_threshold

        print("Label Threshold : ", self.CONFIDENCE_THRESHOLD)

        # Numbers of labels to show
        self.NUM_OF_LABELS = 3

    def has_data(self) -> bool:
        """Checks whethere label data is avalible in the input data.

        :return: [description]
        :rtype: bool
        """

        if self.actions.LABELS in self.input_data.keys():
            labels = self.input_data[self.actions.LABELS]

            if len(labels) > 0:
                return True
            else:
                return False
        else:
            return False

    def apply(self, interim_result: dict) -> dict:
        """Process intermin result from previous rules in the chain and run label rule.

        :param interim_result: Intermediate results from previous rules in the chain.
        :type interim_result: dict
        :return: [description]
        :rtype: dict
        """

        ocr_data_present = len(interim_result[self.actions.OCR]) > 0
        caption_data_present = len(interim_result[self.actions.DESCRIBE]) > 0

        if self.has_data():

            labels_data = self.input_data[self.actions.LABELS]
            filtered_labels = list(filter(lambda label: label["confidence"] >= self.CONFIDENCE_THRESHOLD,
                                          labels_data))

            if GenerationConfig.DEBUG:
                labels = list(map(lambda label: {"description": label["description"], "confidence": label["confidence"]},
                                  filtered_labels[:self.NUM_OF_LABELS]))
            else:
                labels = ", ".join(list(map(lambda label: label["description"], filtered_labels[:self.NUM_OF_LABELS])))
            interim_result[self.actions.LABELS] = labels

            if not ocr_data_present and not caption_data_present:

                if GenerationConfig.DEBUG:
                    interim_result[self.alt_recommendation_key] = interim_result[
                                                                      self.alt_recommendation_key] + ",".join([label["description"] for label in labels]) + "."
                else:
                    interim_result[self.alt_recommendation_key] = interim_result[self.alt_recommendation_key] + labels + "."

        else:
            interim_result[self.actions.LABELS] = ""

        return interim_result
